/*===================================================================*/
/*                                                                   */
/*  pNesX_System_Psx.cpp : The function which depends on a system    */
/*                         (for PSX)                                 */
/*                                                                   */
/*  1999/12/26  Racoon  New preparation                              */
/*  2000/06/07  Racoon                                               */
/*                                                                   */
/*===================================================================*/

/*-------------------------------------------------------------------*/
/*  Include files                                                    */
/*-------------------------------------------------------------------*/

#include "pNesX.h"
#include "pNesX_System.h"
#include "pNesX_Mapper.h"

/*-------------------------------------------------------------------*/
/*  Include files (for PSX)                                          */
/*-------------------------------------------------------------------*/

#include <libps.h>
#include <sys/file.h>
#include <string.h>
#include <memory.h>

/*-------------------------------------------------------------------*/
/*  Definitions                                                      */
/*-------------------------------------------------------------------*/

/* NES screen size */
#define NES_SCREEN_WIDTH    256
#define NES_SCREEN_HEIGHT   240

/* Menu screen size */
#define MENU_SCREEN_WIDTH   640
#define MENU_SCREEN_HEIGHT  480

/* Current screen */
#define SCREEN_MENU 0
#define SCREEN_NES  1

/* Color of menu back ground */
#define MENU_BG_R 255
#define MENU_BG_G 181
#define MENU_BG_B 49

/* Color of selection */
#define SELECT_R  10
#define SELECT_G  87
#define SELECT_B  104

/* Retry count for CD */
#define CD_RETRY_COUNT  10

/* Ordering table size */
#define OT_LENGTH  2

/* Work area */
#define WORK_ADDR  0x80110000

#define MISC_ADDR   WORK_ADDR
#define ASCII_ADDR  ( WORK_ADDR + 0x8000 )
#define ICON_ADDR   ( WORK_ADDR + 0xf000 )
#define VAG_SQUARE_ADDR   (u_long *)( WORK_ADDR + 0x10000 )
#define VAG_TRIANGLE_ADDR (u_long *)( WORK_ADDR + 0x14000 )
#define VAG_NOISE_ADDR    (u_long *)( WORK_ADDR + 0x18000 )


/* ROM address */
#define ROM_ADDR   0x80120000

/* Heap address */
#define HEAP_ADDR  0x801f0000

/* The data storage position of the number for the title */
#define NUMBER_X  640
#define NUMBER_Y  0

/* The CLUT storage position of the number for the title */
#define NUMBER_CLUT_X  880
#define NUMBER_CLUT_Y  0

/* A data storage position for the kanji display */
#define KANJI_X 0
#define KANJI_Y 480

/* A CLUT storage position for the kanji display */
#define KANJI_CLUT_X 0
#define KANJI_CLUT_Y 496

/* The data storage position of the ASCII character */
#define ASCII_X  640
#define ASCII_Y  256

/* The CLUT storage position of the ASCII character */
#define ASCII_CLUT_X  NUMBER_CLUT_X
#define ASCII_CLUT_Y  NUMBER_CLUT_Y + 32

/* The number of the ROM list maximums */
#define ROMLIST_MAX  500

/* Controller state */
#define PAD_STATE  ( ~( *( bb0 + 3 )       | *( bb0 + 2 ) << 8 | \
                        *( bb1 + 3 ) << 16 | *( bb1 + 2 ) << 24 ) )

/* Controller 1 */
#define P1_L2        0x1
#define P1_R2        0x2
#define P1_L1        0x4
#define P1_R1        0x8
#define P1_TRIANGLE  0x10
#define P1_R_UP      0x10
#define P1_CIRCLE    0x20
#define P1_R_RIGHT   0x20
#define P1_CROSS     0x40
#define P1_R_DOWN    0x40
#define P1_SQUARE    0x80
#define P1_R_LEFT    0x80
#define P1_SELECT    0x100
#define P1_I         0x200
#define P1_J         0x400
#define P1_START     0x800
#define P1_UP        0x1000
#define P1_RIGHT     0x2000
#define P1_DOWN      0x4000
#define P1_LEFT      0x8000

/* Controller 2 */
#define P2_L2        0x10000
#define P2_R2        0x20000
#define P2_L1        0x40000
#define P2_R1        0x80000
#define P2_TRIANGLE  0x100000
#define P2_R_UP      0x100000
#define P2_CIRCLE    0x200000
#define P2_R_RIGHT   0x200000
#define P2_CROSS     0x400000
#define P2_R_DOWN    0x400000
#define P2_SQUARE    0x800000
#define P2_R_LEFT    0x800000
#define P2_SELECT    0x1000000
#define P2_I         0x2000000
#define P2_J         0x4000000
#define P2_START     0x8000000
#define P2_UP        0x10000000
#define P2_RIGHT     0x20000000
#define P2_DOWN      0x40000000
#define P2_LEFT      0x80000000

/* Examine whether it is a kanji */
#define IS_KANJI(a)  ( ( (a) >= 0x81 && (a) <= 0x9f ) || ( (a) >= 0xe0 && (a) <= 0xfc ) )

/* Memory card infomation */
struct card_tag{
  char  Magic[2];
  char  Type;
  char  BlockEntry;
  char  Title[64];
  char  reserve[28];
  char  Clut[32];
  char  Icon[3][128];
};

/*-------------------------------------------------------------------*/
/*  Messages                                                         */
/*-------------------------------------------------------------------*/

#define MSG1 "There is no space in the memory card"
#define MSG2 "Saving data aren't kept"
#define MSG3 "Now Loading"
#define MSG4 "It failed in reading"
#define MSG5 "Or, it is the mapper which doesn't comply with it"

/*-------------------------------------------------------------------*/
/*  Global variables                                                 */
/*-------------------------------------------------------------------*/


// Palette data
WORD NesPalette[ 64 ] =
{
  0x39ce, 0x4464, 0x5400, 0x4c08, 0x3811, 0x0815, 0x0014, 0x002f,
  0x00a8, 0x0100, 0x0140, 0x08e0, 0x2ce3, 0x0000, 0x0000, 0x0000,
  0x5ef7, 0x75c0, 0x74e4, 0x7810, 0x5c17, 0x2c1c, 0x00bb, 0x0539,
  0x01d1, 0x0240, 0x02a0, 0x1e40, 0x4600, 0x0000, 0x0000, 0x0000,
  0x7fff, 0x7ee7, 0x7e4b, 0x7e28, 0x7dfe, 0x59df, 0x31df, 0x1e7f,
  0x1efe, 0x0b50, 0x2769, 0x4feb, 0x6fa0, 0x0000, 0x0000, 0x0000,
  0x7fff, 0x7f95, 0x7f58, 0x7f3a, 0x7f1f, 0x6f1f, 0x5aff, 0x577f,
  0x539f, 0x53fc, 0x5fd5, 0x67f6, 0x7bf3, 0x0000, 0x0000, 0x0000
};

/* Pointer of controller buffer */
volatile u_char *bb0, *bb1;

/* Ordering table header */
GsOT ot[ 2 ];
GsOT *pOT;

/* Ordering table */
GsOT_TAG otTags[ 2 ][ 1 << OT_LENGTH ];

/* Packet (How many should I make size?) */
PACKET packet[ 2 ][ 1500 * ( 20 + 4 ) ];

int nActiveBuf;

/* Cureen screen */
int g_nScreen;  /* 0:Menu  1:NES */

/* ROM image infomation */
typedef struct ROM_LIST_TAG
{
  char szTitle[ 39 ];
  char szFileName[ 40 ];
  char chFrameIRQ;
} ROM_LIST;

/* ROM list */
ROM_LIST romlist[ ROMLIST_MAX ];

/* The number of the ROM */
int g_nRomCnt;

/* display start position */
int g_nDispTop;

/* Current select */
int g_nSel;

/* SRAM file name */
char szSaveName[ 256 ];

/* SRAM save flag */
int nSRAM_SaveFlag;

/* ROM file name */
char szRomName[ 48 ];

/* ROM title */
char szTitle[ 48 ];

/* Frame IRQ mode */
char chFrameIRQ;

/* Icon data */
u_char icon_image[ 128 ];

/* Icon CLUT */
u_short icon_clut[ 16 ];  

/*-------------------------------------------------------------------*/
/*  Function prototypes                                              */
/*-------------------------------------------------------------------*/
static void pNesX_Title();

static int LoadSRAM();
static int SaveSRAM();
static int CreateSRAM();

static int ReadFile( const char *pszFileName, void *pBuf, int nSize );

static void ReadRomList();

static void SetResolution( int nScreen, int nClear, BYTE byRed, BYTE byGreen, BYTE byBlue );

static char *mgets( char *pBuf, int nSize, char *pMem );

static void DrawNumbers( char *pszNumbers, int nX, int nY, int nPal );
static void TextOut( int nX, int nY, char *pszText, int nPal, int nInit );

/*===================================================================*/
/*                                                                   */
/*                          main() : Main                            */
/*                                                                   */
/*===================================================================*/
int main()
{
/*
 *  Main
 *
 *  Return values
 *    It will be never returned.
 */

  GsIMAGE imgTemp;
  u_long *pTim;
  int nIdx;
  
  /*-------------------------------------------------------------------*/
  /*  Initialize system                                                */
  /*-------------------------------------------------------------------*/

  // Display off
  SetDispMask( 0 );

  // Reset a graphic system
  ResetGraph( 0 );

  // It doesn't do V-Sync call back
  VSyncCallback( NULL );
  
  // Change the resolution
  SetResolution( SCREEN_MENU, 1, 0, 0, 0 );

  // Display on
  SetDispMask( 1 );

  // Enable interrupt
  ExitCriticalSection();

  // Set the heap area
  InitHeap( (void *)HEAP_ADDR, 0xff00 );

  // Get the pointer of controller buffer
  GetPadBuf( &bb0, &bb1 );

  // Initialize the ordering table
  ot[ 0 ].length = ot[ 1 ].length = OT_LENGTH;
  ot[ 0 ].org = otTags[ 0 ];
  ot[ 1 ].org = otTags[ 1 ];

  // Read the number for the title
  pTim = (u_long *)MISC_ADDR;
  ++pTim;
  GsGetTimInfo( pTim, &imgTemp );
  imgTemp.px = NUMBER_X;
  imgTemp.py = NUMBER_Y;
  LoadImage( (RECT *)&imgTemp.px, imgTemp.pixel );
  imgTemp.cx = NUMBER_CLUT_X;
  imgTemp.cy = NUMBER_CLUT_Y;
  LoadImage( (RECT *)&imgTemp.cx, imgTemp.clut );
  DrawSync( 0 );

  // Read the ascii character
  pTim = (u_long *)ASCII_ADDR;
  ++pTim;
  GsGetTimInfo( pTim, &imgTemp );
  imgTemp.px = ASCII_X;
  imgTemp.py = ASCII_Y;
  LoadImage( (RECT *)&imgTemp.px, imgTemp.pixel );
  imgTemp.cx = ASCII_CLUT_X;
  imgTemp.cy = ASCII_CLUT_Y;
  LoadImage( (RECT *)&imgTemp.cx, imgTemp.clut );
  DrawSync( 0 );

  // Read the ascii character
  pTim = (u_long *)ASCII_ADDR;
  ++pTim;
  GsGetTimInfo( pTim, &imgTemp );
  imgTemp.px = ASCII_X;
  imgTemp.py = ASCII_Y;
  LoadImage( (RECT *)&imgTemp.px, imgTemp.pixel );
  imgTemp.cx = ASCII_CLUT_X;
  imgTemp.cy = ASCII_CLUT_Y;
  LoadImage( (RECT *)&imgTemp.cx, imgTemp.clut );
  DrawSync( 0 );

  // Read Icon data
  pTim = (u_long *)ICON_ADDR;
  ++pTim;
  GsGetTimInfo( pTim, &imgTemp );
  memcpy( icon_image, imgTemp.pixel, 128 );
  memcpy( icon_clut, imgTemp.clut, 32 );

  pNesX_InitSound();
  
  /*-------------------------------------------------------------------*/
  /*  Read the rom list                                                */
  /*-------------------------------------------------------------------*/
  ReadRomList();

  /*-------------------------------------------------------------------*/
  /*  Title animation                                                  */
  /*-------------------------------------------------------------------*/

  pNesX_Title();

  /*-------------------------------------------------------------------*/
  /*  Start the pNesX                                                  */
  /*-------------------------------------------------------------------*/

  pNesX_Main();

  return 0;
}

/*===================================================================*/
/*                                                                   */
/*                pNesX_Title() : Title animation                    */
/*                                                                   */
/*===================================================================*/
static void pNesX_Title()
{
/*
 *  Title animation
 *
 */

  int nTime;
  int nIdx;
  int nState;
  
  int nLogoTime;
  GsLINE line;
  GsSPRITE spr;
  DWORD dwPad1, dwPad2, dwSystem;
  
  struct ylist_tag
  {
    int nState;
    int nDelay;
    char szYear[ 5 ];
    int nX;
    int nY;
    int nPal;
    int nDispTime;
  };
  
  struct ylist_tag ylist[] =
  {
    { 0,   0, "2000", 210, 210, 14, 20 },
    { 0,  32, "1999",  20,  40, 14, 18 },
    { 0,  56, "1998", 100, 240, 14, 16 },
    { 0,  76, "1997", 400, 260, 14, 13 },
    { 0,  94, "1996", 250, 190, 14, 10 },
    { 0, 111, "1995",  30, 300, 14, 10 },
    { 0, 127, "1994", 150,  30, 14, 10 },
    { 0, 142, "1993", 120, 200, 14, 10 },
    { 0, 156, "1992",  80,  80, 14, 10 },
    { 0, 169, "1991", 420,  40, 14, 10 },
    { 0, 181, "1990", 300, 320, 14, 10 },
    { 0, 192, "1989", 240, 200, 14, 10 },
    { 0, 202, "1988", 460, 380, 14, 10 },
    { 0, 211, "1987", 200,  20, 14, 10 },
    { 0, 221, "1986",  70,  80, 14, 13 },
    { 0, 230, "1985", 190, 290, 14, 16 },
    { 0, 239, "1984", 340, 170, 14, 18 },
    { 0, 280, "1983", 210, 210, 14, 20 }
  };
  
  // to Menu screen
  SetResolution( SCREEN_MENU, 1, 0, 0, 0 );
  
  nTime = 0;
  nState = 1;
  nLogoTime = 0;
  
  // Wait the V-Sync
  VSync( 0 );

  nActiveBuf = GsGetActiveBuff();
  pOT = &ot[ nActiveBuf ];

  // Set up the packet area
  GsSetWorkBase( packet[ nActiveBuf ] );

  // Clear the ordering table
  GsClearOt( 0, 0, pOT );

  while ( nState )
  {
    // Cancel the animation
    pNesX_PadState( &dwPad1, &dwPad2, &dwSystem );
    if ( PAD_PUSH( dwSystem, PAD_SYS_OK ) )
      break;

    ++nTime;
    if ( nTime == 6 )
    {
      nTime = 0;

      nActiveBuf = 1 - nActiveBuf;
      pOT = &ot[ nActiveBuf ];

      // Set up the packet area
      GsSetWorkBase( packet[ nActiveBuf ] );

      // Clear the ordering table
      GsClearOt( 0, 0, pOT );

      /*-------------------------------------------------------------------*/
      /*  Display the year                                                 */
      /*-------------------------------------------------------------------*/
      for ( nIdx = 17; nIdx >= 0; --nIdx )
      {
        if ( ylist[ nIdx ].nState > 0 && ylist[ nIdx ].nState < 4 )
        {
          DrawNumbers( ylist[ nIdx ].szYear, ylist[ nIdx ].nX, ylist[ nIdx ].nY, 
                       ylist[ nIdx ].nPal );
        }

        switch ( ylist[ nIdx ].nState )
        {
          case 0:
            if ( ylist[ nIdx ].nDelay == 0 )
              ylist[ nIdx ].nState = 1;
            else
              --ylist[ nIdx ].nDelay;
            break;
            
          case 1:
            --ylist[ nIdx ].nPal;
            if ( ylist[ nIdx ].nPal == 0 )
              ylist[ nIdx ].nState = 2;
            break;
        
          case 2:
            --ylist[ nIdx ].nDispTime;
            if ( ylist[ nIdx ].nDispTime == 0 )
              ylist[ nIdx ].nState = 3;
            break;
          
          case 3:
            ++ylist[ nIdx ].nPal;
            if ( ylist[ nIdx ].nPal == 15 )
              ylist[ nIdx ].nState = 4;
            break;
            
          case 4:
            if ( nIdx == 17 )
              nState = 0;
            break;
        }
      }
    } 

    // Wait the V-Sync
    VSync( 0 );
    ResetGraph( 1 );
    GsSwapDispBuff();

    // Draw the ordering table
    GsDrawOt( pOT );
  }

  ResetGraph( 1 );
}

/*===================================================================*/
/*                                                                   */
/*              SetResolution() : Change the resolution              */
/*                                                                   */
/*===================================================================*/
static void SetResolution( int nScreen, int nClear, BYTE byRed, BYTE byGreen, BYTE byBlue )
{
/*
 *  Change the resolution
 *
 *  Parameters
 *    int nScreen               (Read)
 *      0 : Menu screen
 *      1 : NES screen
 *
 *    int nClear                (Read)
 *      0 : It isn't cleared
 *      1 : Clear it by the specification color
 *
 *    BYTE byRed                (Read)
 *    BYTE byGreen
 *    BYTE byBlue
 *      Clear color
 *      
 */

  static int nPrevScreen = -1;
  RECT rc;

  if ( nPrevScreen != nScreen )
  {
    ResetGraph( 0 );

    if ( nScreen == SCREEN_MENU )
    {
      // Menu screen
      GsInitGraph( MENU_SCREEN_WIDTH, MENU_SCREEN_HEIGHT, 2, 1, 0 ); 
      GsDefDispBuff( 0, 0, 0, 0 );
    }
    else
    {
      // NES screen
      GsInitGraph( NES_SCREEN_WIDTH, NES_SCREEN_HEIGHT, 4, 0, 0 ); 
      GsDefDispBuff( 0, 0, 0, NES_SCREEN_HEIGHT );
    }
  }

  if ( nClear )
  {
    rc.x = 0;
    rc.y = 0;
    if ( nScreen == SCREEN_MENU )
    {
      rc.w = MENU_SCREEN_WIDTH;
      rc.h = MENU_SCREEN_HEIGHT;
    }
    else
    {
      rc.w = NES_SCREEN_WIDTH;
      rc.h = NES_SCREEN_HEIGHT * 2;
    }

    VSync( 0 );
    ClearImage( &rc, byRed, byGreen, byBlue );
    DrawSync( 0 );
    VSync( 0 );
    GsSwapDispBuff();

    ClearImage( &rc, byRed, byGreen, byBlue );
    DrawSync( 0 );
    VSync( 0 );
    GsSwapDispBuff();
  }

  nPrevScreen = nScreen;
}

/*===================================================================*/
/*                                                                   */
/*          DrawNumbers() : Draw the numbers of the title            */
/*                                                                   */
/*===================================================================*/
static void DrawNumbers( char *pszNumbers, int nX, int nY, int nPal )
{
/*
 *  Draw the numbers of the title
 *
 *  Parameters
 *    char *pszNumbers                (Read)
 *      String of the numbers
 *
 *    int nX                          (Read)
 *      X position
 *
 *    int nY                          (Read)
 *      Y position
 *
 *    int nPal                        (Read)
 *      Number of palette
 *
 */
 
  GsSPRITE spr;
  int nIdx;
  int nNum;

  spr.attribute = 0;
  spr.w = 44;
  spr.h = 50; 
  spr.tpage = GetTPage( 0, 0, NUMBER_X, NUMBER_Y );
  spr.cx = NUMBER_CLUT_X;
  spr.cy = NUMBER_CLUT_Y + nPal;
  spr.r = spr.g = spr.b = 0x80;
  spr.mx = 0;
  spr.my = 0;
  spr.scalex = ONE;
  spr.scaley = ONE;
  spr.rotate = 0;
    
  spr.y = nY;

  // Draw it each one character
  for ( nIdx = 0; pszNumbers[ nIdx ] != '\0'; ++nIdx )
  {
    nNum = pszNumbers[ nIdx ] - '0';
    spr.x = nX + nIdx * 44;
    spr.u = ( nNum % 5 ) * 44;
    spr.v = ( nNum / 5 ) * 51;

    GsSortFastSprite( &spr, pOT, 0 );
  }
}

/*===================================================================*/
/*                                                                   */
/*                    TextOut() : Draw a text                        */
/*                                                                   */
/*===================================================================*/
static void TextOut( int nX, int nY, char *pszText, int nPal, int nInit )
{
/*
 *  Draw a text
 *
 *  Parameters
 *   int nX                          (Read)
 *     X position
 *
 *   int nY                          (Read)
 *     Y position
 *
 *   char *pszText                   (Read)
 *     Text
 *
 *   int nPal                        (Read)
 *     Number of palette
 *
 *   int nInit                       (Read)
 *       1 : The initialization of the kanji data preparation position
 *       0 : It isn't initialized
 *
 */

  u_long *pTim;
  GsSPRITE spr;
  GsIMAGE imgStr;
  RECT rc;
  int nIdx;
  u_char pszKanji[ 257 ];
  int nPos;
  static int nKanjiPos;
  u_char uchAscii;
  int nKanjiLen;
  
  if ( nInit )
    nKanjiPos = 0;  // The initialization of the kanji data preparation position
  
  // CLUT preparation for the kanji
  rc.x = ASCII_CLUT_X;
  rc.y = ASCII_CLUT_Y + nPal;
  rc.w = 1;
  rc.h = 1;
  MoveImage( &rc, KANJI_CLUT_X, KANJI_CLUT_Y + nPal );
  rc.x = ASCII_CLUT_X + 15;
  MoveImage( &rc, KANJI_CLUT_X + 1, KANJI_CLUT_Y + nPal );
  
  // Set up a sprite
  spr.attribute = 0;
  spr.h = 16;
  spr.r = spr.g = spr.b = 0x80;
  spr.mx = 0;
  spr.my = 0;
  spr.scalex = ONE;
  spr.scaley = ONE;
  spr.rotate = 0;

  // Draw it each one character
  nIdx = 0;
  nPos = 0;
  while ( pszText[ nIdx ] != '\0' )
  {
    if ( IS_KANJI( (u_char)pszText[ nIdx ] ) )
    {
      // Kanji
      
      // Find the number that a kanji continues. (In texture page)
      for ( nKanjiLen = 1; ( ( nKanjiPos + nKanjiLen * 4 ) & 0x3f ) &&
                           IS_KANJI( (u_char)pszText[ nIdx + nKanjiLen * 2 ] ); ++nKanjiLen )
        ;
      
      // Make the TIM image of the text in the main memory
      pTim = (u_long *)WORK_ADDR;
      memcpy( pszKanji, pszText + nIdx, nKanjiLen * 2 );
      pszKanji[ nKanjiLen * 2 ] = '\0';
      Krom2Tim( pszKanji, pTim, KANJI_X, KANJI_Y, KANJI_CLUT_X, KANJI_CLUT_Y, 0x00ffffff, 0x00000000 );
  
      // Transfer to the frame buffer
      GsGetTimInfo( pTim + 1, &imgStr );
      imgStr.px += nKanjiPos;
      LoadImage( (RECT *)&imgStr.px, imgStr.pixel );
      DrawSync( 0 );
      
      // Set up a sprite
      spr.x = nX + nPos;
      spr.y = nY;
      spr.w = nKanjiLen * 16;
      spr.tpage = GetTPage( 0, 0, KANJI_X + ( nKanjiPos & 0xffc0 ), KANJI_Y & 0x100 );
      spr.u = ( nKanjiPos & 0x3f ) * 4;
      spr.v = KANJI_Y - 256;
      spr.cx = KANJI_CLUT_X;
      spr.cy = KANJI_CLUT_Y + nPal;

      GsSortFastSprite( &spr, pOT, 0 );

      ++spr.x;
      GsSortFastSprite( &spr, pOT, 0 );
      
      nIdx += nKanjiLen * 2;
      nPos += nKanjiLen * 16;
      nKanjiPos += nKanjiLen * 4;      
    }
    else
    {
      // Ascii character
      
      uchAscii = (u_char)pszText[ nIdx ];
      uchAscii = ( uchAscii >= 32 && uchAscii < 128 ) ? uchAscii - 0x20 : 0;
      
      // Set up a sprite
      spr.x = nX + nPos;
      spr.y = nY + 2;
      spr.w = 8;
      spr.tpage = GetTPage( 0, 0, ASCII_X, ASCII_Y );
      spr.u = ( uchAscii % 32 ) * 8;
      spr.v = ( uchAscii / 32 ) * 16;
      spr.cx = ASCII_CLUT_X;
      spr.cy = ASCII_CLUT_Y + nPal;

      GsSortFastSprite( &spr, pOT, 0 );
      
      ++nIdx;
      nPos += 8;
    }
  }
}

/*===================================================================*/
/*                                                                   */
/*                ReadRomList() : Read the rom list                  */
/*                                                                   */
/*===================================================================*/
static void ReadRomList()
{
/*
 *  Read the rom list
 *
 */

  char *pBuf;
  int nSize;
  int nTag = 0;
  int nIdx;
  char szLine[ 257 ];
  char szTag[ 257 ];
  char *pszItem;
  char *pszValue;
  char *pMem;

  g_nRomCnt = 0;
  pMem = (char *)WORK_ADDR;
  nSize = ReadFile( "\\ROMLIST.INI;1", pMem, 0 );
  if ( nSize > 0 )
  {
    pMem[ nSize ] = '\0';
  
    // Analyze it respectively one line
    while ( mgets( szLine, sizeof szLine, pMem ) )
    {
      // Skip a comment line
      if ( strchr( "!#;'\r\n", szLine[ 0 ] ) )
        continue;

      if ( szLine[ 0 ] == '[' )
      {
        // Tag
        if ( nTag )
          ++g_nRomCnt;
        
        strcpy( szTag, &szLine[ 1 ] );
        
        // Remove an unnecessary character
        for ( nIdx = strlen( szTag ) - 1; nIdx >= 0; --nIdx )
        {
          if ( strchr( " ]\r\n", szTag[ nIdx ] ) == NULL )
            break;
        }
        szTag[ nIdx + 1 ] = '\0';
        
        // Make tag the title of the ROM
        memset( &romlist[ g_nRomCnt ], sizeof( ROM_LIST ), 0 );
        strcpy( romlist[ g_nRomCnt ].szTitle, szTag );
        nTag = 1;
      }
      else
      {
        // Parameters
        if ( nTag )
        {
          // Get a item name
          pszItem = strtok( szLine, "=" );
          // Get a item value
          pszValue = strtok( NULL, "\r\n" );
          
          if ( pszItem && pszValue )
          {
            // Set a value
            if ( strcmp( pszItem, "FileName" ) == 0 )
            {
              // File name
              strcpy( romlist[ g_nRomCnt ].szFileName, pszValue );
            }
            else
            if ( strcmp( pszItem, "FrameIRQ" ) == 0 )
            {
              // IRQ is made to occur at the time of V-Sync.
              romlist[ g_nRomCnt ].chFrameIRQ = atoi( pszValue );
            }

          }
        }
      } // if ( szLine[ 0 ] == '[' )
      
    } // while ( mgets ... 

    // The count of the last
    if ( nTag )
      ++g_nRomCnt;
  }
}

/*===================================================================*/
/*                                                                   */
/*              mgets() : Get a one line from the memory             */
/*                                                                   */
/*===================================================================*/
static char *mgets( char *pBuf, int nSize, char *pMem )
{
/*
 *  Get a one line from the memory
 *
 *  Parameters
 *    char *pBuf                      (Read)
 *      Pointer to the buffer which a string is stored in
 *
 *    int nSize                       (Read)
 *      Size of buffer
 *
 *    char *pMem                      (Read)
 *      Pointer to the target data
 *
 *  Return value
 *    Pointer to string.
 *    Return NULL when a string doesn't exist any further.
 *
 *  Remarks
 *    The next line can be acquired by delivering the value 
 *    which is the same as the last time to pMEM.
 */
 
  static char *pMemStart;
  static char *pMemPos;
  int nLen;
  
  if ( pMem == pMemStart )
  {
    // Continuation search
    if ( pMemPos && *pMemPos == '\0' )
    {
      // Search completion
      pMemStart = NULL;
      return NULL;
    }
    
    pMem = pMemPos;
  }
  else
  {
    // Search start
    pMemStart = pMem;
    pMemPos = pMem;
  }
  
  // Until a NULL character or a line feed code appears, loop
  for ( nLen = 0; pMem[ nLen ] != '\0' && pMem[ nLen ] != '\r' && pMem[ nLen ] != '\n'; ++nLen )
    ;
  
  if ( pMem[ nLen ] != '\0' )
  {
    // A line feed code appeared
    if ( pMem[ nLen ] == '\r' && pMem[ nLen + 1 ] == '\n' )
      nLen += 2; // DOS
    else
      ++nLen; // Others
  }
  
  // Search start position movement
  pMemPos += nLen;
  
  // Set up an acquisition character line head
  if ( nLen > nSize - 1 )
    nLen = nSize - 1;
  
  // Copy a string on the buffer
  memcpy( pBuf, pMem, nLen );
  pBuf[ nLen ] = '\0';
  
  // Return a pointer to the string
  return pMem;
}

/*===================================================================*/
/*                                                                   */
/*                ReadFile() : Read data from the CD                 */
/*                                                                   */
/*===================================================================*/
static int ReadFile( const char *pszFileName, void *pBuf, int nSize )
{
/*
 *  Read data from the CD
 *
 *  Parameters
 *    char *pszFileName                      (Read)
 *      FileName
 *
 *    void *pBuf                             (Read)
 *      Pointer to the reading buffer
 *
 *    int nSize                              (Read)
 *      Reading size
 *      When 0 is specified, for the file size
 *
 *  Return values
 *   >=0 : Read size
 *    -1 : Error
 */

  int nRetry;          // Retry counts
  CdlFILE cdl, *pcdl;  // Information of the file

  // Get the file size
  for ( nRetry = 0; nRetry < CD_RETRY_COUNT; ++nRetry )
  {
    pcdl = CdSearchFile( &cdl, (char *)pszFileName );

    // It goes through the loop if file size can be acquired
    if ( !( pcdl == (CdlFILE *)0 || pcdl == (CdlFILE *)-1 ) )
      break;

    VSync( 60 );  // Wait
  }
  if ( nRetry == CD_RETRY_COUNT )
    return -1;  // Error

  // Calculate size to read
  if ( nSize == 0 || nSize > cdl.size )
    nSize = cdl.size;

  // Retry loop
  for ( nRetry = 0; nRetry < CD_RETRY_COUNT; ++nRetry )
  {
    // Read a data
    CdReadFile( (char *)pszFileName, (u_long *)pBuf, nSize );

    // Wait for reading
    if ( CdReadSync( 0, 0 ) == 0 )
      return nSize;  // Successful

    VSync( 60 ); // Wait
  }

  // Error
  return -1;
}

/*===================================================================*/
/*                                                                   */
/*           LoadSRAM() : Load a SRAM                                */
/*                                                                   */
/*===================================================================*/
static int LoadSRAM()
{
/*
 *  Load a SRAM
 *
 *  Return values
 *     0 : Normally
 *    -1 : SRAM data couldn't be read
 */
  long fd;
  unsigned char pSrcBuf[ SRAM_SIZE ];
  unsigned char chData;
  unsigned char chTag;
  int nRunLen;
  int nDecoded;
  int nDecLen;
  int nIdx;
  char *pName;

  // It doesn't need to save it
  nSRAM_SaveFlag = 0;

  // It is finished if the ROM doesn't have SRAM
  if ( !ROM_SRAM )
    return 0;

  // There is necessity to save it
  nSRAM_SaveFlag = 1;

  // The preparation of the SRAM file name
  pName = strrchr( szRomName, '\\' );
  if ( pName == NULL )
    pName = szRomName;
  strcpy( szSaveName, "bu00:" );
  strcat( szSaveName, pName );
  strcpy( strrchr( szSaveName, '.' ) + 1, "srm" );

  /*-------------------------------------------------------------------*/
  /*  Read a SRAM data                                                 */
  /*-------------------------------------------------------------------*/

  // Open SRAM file
  fd = open( szSaveName, O_RDONLY );
  if ( fd < 0 )
    return -1;

  // Read SRAM data
  lseek( fd, 256, SEEK_SET );
  read( fd, (char *)pSrcBuf, SRAM_SIZE - 256 );

  // Close SRAM file
  close( fd );

  /*-------------------------------------------------------------------*/
  /*  Extract a SRAM data                                              */
  /*-------------------------------------------------------------------*/

  nDecoded = 0;
  nDecLen = 0;

  chTag = pSrcBuf[ nDecoded++ ];

  while ( nDecLen < 8192 )
  {
    chData = pSrcBuf[ nDecoded++ ];

    if ( chData == chTag )
    {
      chData = pSrcBuf[ nDecoded++ ];
      nRunLen = pSrcBuf[ nDecoded++ ];
      for ( nIdx = 0; nIdx < nRunLen + 1; ++nIdx )
      {
        SRAM[ nDecLen++ ] = chData;
      }
    }
    else
    {
      SRAM[ nDecLen++ ] = chData;
    }
  }

  // Successful
  return 0;
}

/*===================================================================*/
/*                                                                   */
/*           SaveSRAM() : Save a SRAM                                */
/*                                                                   */
/*===================================================================*/
static int SaveSRAM()
{
/*
 *  Save a SRAM
 *
 *  Return values
 *     0 : Normally
 *    -1 : SRAM data couldn't be written
 */

  long fd;
  struct card_tag card;
  int nUsedTable[ 256 ];
  unsigned char chData;
  unsigned char chPrevData;
  unsigned char chTag;
  int nIdx;
  int nEncoded;
  int nEncLen;
  int nRunLen;
  unsigned char pDstBuf[ SRAM_SIZE ];

  if ( !nSRAM_SaveFlag )
    return 0;  // It doesn't need to save it

  /*-------------------------------------------------------------------*/
  /*  Compress a SRAM data                                             */
  /*-------------------------------------------------------------------*/

  memset( nUsedTable, 0, sizeof nUsedTable );

  for ( nIdx = 0; nIdx < SRAM_SIZE; ++nIdx )
  {
    ++nUsedTable[ SRAM[ nIdx++ ] ];
  }
  for ( nIdx = 1, chTag = 0; nIdx < 256; ++nIdx )
  {
    if ( nUsedTable[ nIdx ] < nUsedTable[ chTag ] )
      chTag = nIdx;
  }

  nEncoded = 0;
  nEncLen = 0;
  nRunLen = 1;

  pDstBuf[ nEncLen++ ] = chTag;

  chPrevData = SRAM[ nEncoded++ ];

  while ( nEncoded < SRAM_SIZE && nEncLen < SRAM_SIZE - 133 )
  {
    chData = SRAM[ nEncoded++ ];

    if ( chPrevData == chData && nRunLen < 256 )
      ++nRunLen;
    else
    {
      if ( nRunLen >= 4 || chPrevData == chTag )
      {
        pDstBuf[ nEncLen++ ] = chTag;
        pDstBuf[ nEncLen++ ] = chPrevData;
        pDstBuf[ nEncLen++ ] = nRunLen - 1;
      }
      else
      {
        for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
          pDstBuf[ nEncLen++ ] = chPrevData;
      }

      chPrevData = chData;
      nRunLen = 1;
    }

  }
  if ( nRunLen >= 4 || chPrevData == chTag )
  {
    pDstBuf[ nEncLen++ ] = chTag;
    pDstBuf[ nEncLen++ ] = chPrevData;
    pDstBuf[ nEncLen++ ] = nRunLen - 1;
  }
  else
  {
    for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
      pDstBuf[ nEncLen++ ] = chPrevData;
  }

  /*-------------------------------------------------------------------*/
  /*  Write a SRAM data                                                */
  /*-------------------------------------------------------------------*/
  
  // Open SRAM file
  fd = open( szSaveName, O_WRONLY );
  if ( fd < 0 )
    return -1;
    
  lseek( fd, 0, SEEK_SET );

  memset( &card, 0, sizeof card );

  card.Magic[ 0 ] = 'S';
  card.Magic[ 1 ] = 'C';
  card.Type = 0x11;
  card.BlockEntry = 1;

  strcpy( card.Title, "pNesX " );
  strcat( card.Title, szTitle );
  memcpy( card.Clut, icon_clut, 32 );
  memcpy( card.Icon[ 0 ], icon_image, 128 );

  // Write block header
  write( fd, (char *)&card, 256 );

  // Write SRAM data
  write( fd, (char *)pDstBuf, nEncLen );

  // Close SRAM file
  close( fd );

  // Successful
  return 0;
}

/*===================================================================*/
/*                                                                   */
/*      CreateSRAM() : Create a SRAM block on the memory card        */
/*                                                                   */
/*===================================================================*/
static int CreateSRAM()
{
/*
 *  Create a SRAM block on the memory card
 *
 *  Return values
 *     0 : Normally
 *    -1 : Error
 */

  long fd;
  struct card_tag card;
  char dummy[ SRAM_SIZE - 256 ];

  // Create a memory block
  fd = open( szSaveName, O_CREAT | ( 1 << 16 ) );
  if ( fd < 0 )
    return -1;
    
  close( fd );

  // Write a header data
  fd = open( szSaveName, O_WRONLY );
  if ( fd < 0 )
    return -1;
    
  lseek( fd, 0, SEEK_SET );

  memset( &card, 0, sizeof card );

  card.Magic[ 0 ] = 'S';
  card.Magic[ 1 ] = 'C';
  card.Type = 0x11;
  card.BlockEntry = 1;

  strcpy( card.Title, "pNesX " );
  strcat( card.Title, szTitle );
  memcpy( card.Clut, icon_clut, 32 );
  memcpy( card.Icon[ 0 ], icon_image, 128 );

  // Write block header
  write( fd, (char *)&card, 256 );

  // Write dummy SRAM data
  memset( dummy, 0, sizeof dummy );
  write( fd, dummy, sizeof dummy );

  // Close SRAM file
  close( fd );

  return 0;
}

/*===================================================================*/
/*                                                                   */
/*                   pNesX_Menu() : Menu screen                      */
/*                                                                   */
/*===================================================================*/
int pNesX_Menu()
{
/*
 *  Menu screen
 *
 *  Return values
 *     0 : Normally
 *    -1 : Exit pNesX (It is never finished on PSX)
 */

  int nPrevSel;
  int nTime;
  int nIdx;
  int nState;
  int nTiming[] = { 2, 2, 2, 2, 2, 60, 60, 2 };
  int nOT;

  int nLogoTime;
  GsSPRITE spr;
  GsBOXF box;
  
  int nLogoPosX[] = { 600, 500, 400, 300, 200, 100,  30,  30,  30,  30, 30, 30, 0 };
  int nLogoPosY[] = { 220, 220, 220, 220, 220, 220, 220, 220, 180, 120, 80, 30, 0 };
  
  int nListIdx;
  DWORD dwPad1, dwPad2, dwSystem;
  DWORD dwPadState;

  // to Menu screen
  SetResolution( SCREEN_MENU, 1, MENU_BG_R, MENU_BG_G, MENU_BG_B );

  // Save a SRAM
  SaveSRAM();

  if ( g_nSel >= g_nRomCnt )
  {
    // New disk
    g_nSel = g_nDispTop = 0;
  }

  nTime = 0;
  nState = 1;
  nLogoTime = 0;
  nPrevSel = -1;

  // Wait the V-Sync
  VSync( 0 );

  nActiveBuf = GsGetActiveBuff();
  pOT = &ot[ nActiveBuf ];

  // Set up the packet area
  GsSetWorkBase( packet[ nActiveBuf ] );

  // Clear the ordering table
  GsClearOt( 0, 0, pOT );

  while ( nState )
  {
    GsDRAWENV.dfe = 1;
    PutDrawEnv( &GsDRAWENV );

    ++nTime;
    if ( nTime == nTiming[ nState - 1 ] )
    {
      nTime = 0;

      nActiveBuf = 1 - nActiveBuf;
      pOT = &ot[ nActiveBuf ];

      // Set up the packet area
      GsSetWorkBase( packet[ nActiveBuf ] );

      // Clear the ordering table
      GsClearOt( 0, 0, pOT );

      switch ( nState )
      {
        case 1:
          /*-------------------------------------------------------------------*/
          /*  Move the pNesX logo                                              */
          /*-------------------------------------------------------------------*/
          if ( nLogoTime == 0 )
          {
            spr.attribute = 0;
            spr.w = 230;
            spr.h = 56;
            spr.tpage = GetTPage( 0, 0, NUMBER_X, NUMBER_Y );
            spr.u = 0;
            spr.v = 102;
            spr.cx = NUMBER_CLUT_X;
            spr.cy = NUMBER_CLUT_Y + 15;
            spr.r = spr.g = spr.b = 0x80;
            spr.mx = 0;
            spr.my = 0;
            spr.scalex = ONE;
            spr.scaley = ONE;
            spr.rotate = 0;
          }
          
          spr.x = nLogoPosX[ nLogoTime ];
          spr.y = nLogoPosY[ nLogoTime ];

          if ( spr.x > 0 )
          {
            GsSortFastSprite( &spr, pOT, 0 );

            if ( nLogoTime > 0 )
            {
              box.attribute = 0;
              box.x = nLogoPosX[ nLogoTime - 1 ];
              box.y = nLogoPosY[ nLogoTime - 1 ];
              box.w = 230;
              box.h = 56;
              box.r = MENU_BG_R;
              box.g = MENU_BG_G;
              box.b = MENU_BG_B;
            
              GsSortBoxFill( &box, pOT, 0 );
            }
          }
          else
          {
            spr.w = 110;
            spr.h = 15;
            spr.u = 0;
            spr.v = 159;
            spr.cy = NUMBER_CLUT_Y + 15;
            spr.x = 500;
            spr.y = 60;

            GsSortFastSprite( &spr, pOT, 0 );

            spr.w = 82;
            spr.h = 22;
            spr.u = 116;
            spr.v = 159;
            spr.cy = NUMBER_CLUT_Y;
            spr.x = 16;
            spr.y = 430;

            GsSortFastSprite( &spr, pOT, 0 );

            spr.w = 200;
            spr.h = 58;
            spr.u = 0;
            spr.v = 181;
            spr.cy = NUMBER_CLUT_Y + 16;
            spr.x = 426;
            spr.y = 412;

            GsSortFastSprite( &spr, pOT, 0 );

            nState = 2;
            nLogoTime = 0;
            break;
          }
          
          ++nLogoTime;
          break;
          
        case 2:
          /*-------------------------------------------------------------------*/
          /*  Draw the window of the rom list                                  */
          /*-------------------------------------------------------------------*/
          if ( nLogoTime == 10 )
          {
            box.attribute = 0;
            box.x = 320;
            box.y = 256;
            box.w = 2;
            box.h = 1;
            box.r = 8;
            box.g = 8;
            box.b = 8;

            GsSortBoxFill( &box, pOT, 0 );
          }
          else
          if ( nLogoTime >= 30 && nLogoTime < 50 )
          {
            box.x -= 16;
            box.w += 32;
          
            GsSortBoxFill( &box, pOT, 0 );
          }
          else
          if ( nLogoTime >= 60 && nLogoTime < 69 )
          {
            box.y -= 16;
            box.h += 32;
          
            GsSortBoxFill( &box, pOT, 0 );
          }
          else
          if ( nLogoTime == 69 )
          {
            nState = 3;
            nLogoTime = 0;
            break;
          }

          ++nLogoTime;
          break;
          
        case 3:
          /*-------------------------------------------------------------------*/
          /*  Display the list                                                 */
          /*-------------------------------------------------------------------*/
          if ( nPrevSel < 0 )
          {
            box.attribute = 0;
            box.x = 149;
            box.y = 136;
            box.w = 16 * 20 + 22;
            box.h = 20 * 12 + 6;
            box.r = MENU_BG_R;
            box.g = MENU_BG_G;
            box.b = MENU_BG_B;

            GsSortBoxFill( &box, pOT, 3 );
          }

          if ( g_nSel != nPrevSel )
          {
            box.x = 152;
            box.y = 139;
            box.w = 16 * 20 + 16;
            box.h = 20 * 12;
            box.r = 8;
            box.g = 8;
            box.b = 8;

            GsSortBoxFill( &box, pOT, 2 );

            TextOut( 0, 0, "", 0, 1 );
          
            for ( nIdx = 0; nIdx < 12; ++nIdx )
            {
              nListIdx = nIdx + g_nDispTop;
              if ( nListIdx < g_nRomCnt )
              {
                if ( nListIdx == g_nSel )
                {
                  TextOut( 160, 140 + nIdx * 20, romlist[ nListIdx ].szTitle, 2, 0 );

                  box.x = 152;
                  box.y = 139 + nIdx * 20;
                  box.w = 16 * 20 + 16;
                  box.h = 20;
                  box.r = SELECT_R;
                  box.g = SELECT_G;
                  box.b = SELECT_B;

                  GsSortBoxFill( &box, pOT, 1 );
                }
                else
                {
                  TextOut( 160, 140 + nIdx * 20, romlist[ nListIdx ].szTitle, 1, 0 );
                }
              }
            }

            nPrevSel = g_nSel;
          }
          
          pNesX_PadState( &dwPad1, &dwPad2, &dwSystem );
          dwPadState = PAD_STATE;
          
          if ( PAD_PUSH( dwSystem, PAD_SYS_OK ) )
          {
            // ROM was selected

            nState = 4;
          }
          else
          if ( PAD_PUSH( dwSystem, PAD_SYS_UP ) )
          {
            // Cursor up
            g_nSel = ( g_nSel > 0 ) ? g_nSel - 1 : 0;
            if ( ( g_nSel - g_nDispTop ) == 1 && g_nDispTop > 0 )
              --g_nDispTop;
          }
          else
          if ( PAD_PUSH( dwSystem, PAD_SYS_DOWN ) )
          {
            // Cursor down
            g_nSel = ( g_nSel < g_nRomCnt - 1 ) ? g_nSel + 1 : g_nRomCnt - 1;
            if ( ( g_nSel - g_nDispTop ) == 10 && g_nDispTop + 12 < g_nRomCnt )
              ++g_nDispTop;
          }
          else
          if ( PAD_PUSH( dwSystem, PAD_SYS_QUIT ) )
          {
            if ( szRomName[ 0 ] != '\0' )
            {
              // to NES mode
              nState = 8;
            }
          }
          else
          if ( PAD_PUSH( dwPadState, P1_R1 ) && PAD_PUSH( dwPadState, P1_R2 ) )
          {
            // Reset
            if ( szRomName[ 0 ] != '\0' )
            {
              // Reset pNesX
              pNesX_Reset();

              // to NES mode
              nState = 8;
            }
          }
          break;

        case 4:
          /*-------------------------------------------------------------------*/
          /*  A ROM was chosen 1                                               */
          /*-------------------------------------------------------------------*/

          // Clear the screen
          box.attribute = 0;
          box.x = 149;
          box.y = 136;
          box.w = 16 * 20 + 22;
          box.h = 20 * 12 + 6;
          box.r = 8;
          box.g = 8;
          box.b = 8;

          GsSortBoxFill( &box, pOT, 2 );

          TextOut( 190, 240, MSG3, 1, 1 );
          TextOut( 190, 270, romlist[ g_nSel ].szTitle, 1, 0 );

          nState = 5;
          break;

        case 5:
          /*-------------------------------------------------------------------*/
          /*  A ROM was chosen 2                                               */
          /*-------------------------------------------------------------------*/

          // Read the ROM image
          if ( pNesX_Load( romlist[ g_nSel ].szFileName ) < 0 )
          {
            // It failed in reading
            nState = 6;
            break;
          }

          /*-------------------------------------------------------------------*/
          /*  Load OK                                                          */
          /*-------------------------------------------------------------------*/

          // Set ROM name and Title
          strcpy( szRomName, romlist[ g_nSel ].szFileName );
          strcpy( szTitle, romlist[ g_nSel ].szTitle );

          // Set Frame IRQ mode
          chFrameIRQ = romlist[ g_nSel ].chFrameIRQ;

          // Load a save data
          if ( LoadSRAM() < 0 )
          {
            // Create a new block
            if ( CreateSRAM() < 0 )
            {
              nState = 7;
              break;
            }
          }

          // to NES mode
          nState = 8;
          break;

        case 6:
          /*-------------------------------------------------------------------*/
          /*  A ROM was chosen 3 (Read error)                                  */
          /*-------------------------------------------------------------------*/

          // Print error message
          if ( nLogoTime == 0 )
          {
            GsSortBoxFill( &box, pOT, 2 );

            TextOut( 140, 240, MSG4, 1, 1 );
            TextOut( 140, 270, MSG5, 1, 0 );
          }
          else
          if ( nLogoTime == 5 )
          {
            // to List mode
            nState = 3;
            nLogoTime = 0;
            nPrevSel = -1;

            box.x = 0;
            box.w = 640;
            GsSortBoxFill( &box, pOT, 0 );
            break;
          }

          ++nLogoTime;
          break;

        case 7:
          /*-------------------------------------------------------------------*/
          /*  A ROM was chosen 4 (Memory card error)                           */
          /*-------------------------------------------------------------------*/

          // Print error message
          if ( nLogoTime == 0 )
          {
            GsSortBoxFill( &box, pOT, 2 );

            TextOut( 170, 240, MSG1, 1, 1 );
            TextOut( 170, 270, MSG2, 1, 0 );
          }
          else
          if ( nLogoTime == 5 )
          {
            // to NES mode
            nState = 8;
            nLogoTime = 0;
            break;
          }

          ++nLogoTime;
          break;

        case 8:
          /*-------------------------------------------------------------------*/
          /*  to NES mode                                                      */
          /*-------------------------------------------------------------------*/
          ResetGraph( 1 );
          
          // Change to the NES screen
          SetResolution( SCREEN_NES, 1, 0, 0, 0 );
 
          // to NES mode
          nState = 0;
          return 0;
      }
    }

    //DrawSync( 0 );
    VSync( 0 );
    ResetGraph( 1 );
    GsSwapDispBuff();

    // Draw the ordering table
    GsDrawOt( pOT );    
  }

  // to NES mode
  return 0;
}

/*===================================================================*/
/*                                                                   */
/*               pNesX_ReadRom() : Read ROM image file               */
/*                                                                   */
/*===================================================================*/
int pNesX_ReadRom( const char *pszFileName )
{
/*
 *  Read ROM image file
 *
 *  Parameters
 *    const char *pszFileName          (Read)
 *
 *  Return values
 *     0 : Normally
 *    -1 : Error
 */

  // Set the ROM address
  ROM = (BYTE *)ROM_ADDR;
  
  // Read the ROM image
  if ( ReadFile( pszFileName, ROM, 0 ) < 0 )
    return -1;
  
  // Read the ROM Header
  memcpy( &NesHeader, ROM, sizeof NesHeader );
  if ( memcmp( NesHeader.byID, "NES\x1a", 4 ) != 0 )
  {
    // not .nes file
    return -1;
  }

  ROM += sizeof NesHeader;

  // Clear SRAM
  pNesX_MemorySet( SRAM, 0, SRAM_SIZE );

  // If trainer presents Read Triner at 0x7000-0x71ff
  if ( NesHeader.byInfo1 & 4 )
  {
    memcpy( &SRAM[ 0x1000 ], ROM, 512 );
    ROM += 512;
  }

  // Set the VROM address
  VROM = ROM + NesHeader.byRomSize * 0x4000;

  // Successful
  return 0;
}

/*===================================================================*/
/*                                                                   */
/*           pNesX_ReleaseRom() : Release a memory for ROM           */
/*                                                                   */
/*===================================================================*/
void pNesX_ReleaseRom()
{
/*
 *  Release a memory for ROM
 *
 */

  // Nothing is done
}

/*===================================================================*/
/*                                                                   */
/*      pNesX_LoadFrame() :                                          */
/*           Transfer the contents of work frame on the screen       */
/*                                                                   */
/*===================================================================*/
void pNesX_LoadFrame()
{
/*
 *  Transfer the contents of work frame on the screen
 *
 */
 
  RECT rc;
  GsSPRITE spr;
  u_short tpage;
  short Active;
  WORD wX, wY;
  BYTE *pbyNameTable;
  BYTE *pSPRRAM;
  WORD wCHR;
  short byXScr, byYScr;
  BYTE byNameTable;
  BYTE *pAttrBase;
  WORD wPX, wPY, wY4;
  GsBOXF box;
  BYTE byBGBase;
  BYTE bySPBase;
  static int    HSTime;
  int attr;
  
  Active = GsGetActiveBuff();
  pOT = &ot[ Active ];

  // Set up the packet area
  GsSetWorkBase( packet[ Active ] );

  // Clear the ordering table
  GsClearOt( 0, 0, pOT );

  box.attribute = 0;
  box.r = 0;
  box.g = 0;
  box.b = 0;
  box.x = 0;
  box.y = 0;

  if ( !( PPU_R1 & R1_SHOW_SCR ) )
  {
    box.w = NES_SCREEN_WIDTH;
    box.h = NES_SCREEN_HEIGHT;
            
    GsSortBoxFill( &box, pOT, 0 );

    DrawSync( 0 );
  
    VSync( 0 );
  
    // Switching of the double buffer
    GsSwapDispBuff();
  
    GsDrawOt( pOT );

    return;
  }

  // h-sync time
  box.y = NES_SCREEN_HEIGHT - 6;
  box.w = HSTime >> 3;
  box.h = 2;

  box.b = 200;

  GsSortBoxFill( &box, pOT, 0 );

  // Clear top and bottom and left
  box.y = 0;
  box.w = NES_SCREEN_WIDTH;
  box.h = 8;
  box.b = 0;
            
  GsSortBoxFill( &box, pOT, 0 );

  box.y = NES_SCREEN_HEIGHT - 8;

  GsSortBoxFill( &box, pOT, 0 );


#if 0
  if ( PPU_R1 & R1_CLIP_SP )
  {
    box.x = 0;
    box.y = 8;
    box.w = 8;
    box.h = NES_SCREEN_HEIGHT - 16;
            
    GsSortBoxFill( &box, pOT, 0 );
  }
#endif

  // Transfer CLUT
  if ( byCLUT_Change )
  {
    rc.x = 640;
    rc.y = 400;
    rc.w = 4;
    rc.h = 8;
    LoadImage( &rc, (u_long *)PalTable );
    byCLUT_Change = 0;
  }
  
  tpage = GetTPage(0, 0, 640 + 256, 256);

  spr.attribute = 0;
  spr.tpage = tpage;
  spr.w = 8;
  spr.h = 8;
  spr.u = 0;
  spr.v = 0;
  spr.cx = 640;
  spr.cy = 400;
  spr.r = spr.g = spr.b = 0x80;
  spr.mx = 0;
  spr.my = 0;
  spr.scalex = ONE;
  spr.scaley = ONE;
  spr.rotate = 0;

  bySPBase = ( ( PPU_R0 & R0_SP_ADDR ) << 4 );
  
  if ( PPU_R1 & R1_SHOW_SP )
  {
    // Reset Scanline Sprite Count
    PPU_R2 &= ~R2_MAX_SP;

    // Render a sprite to the sprite buffer
    //for ( pSPRRAM = SPRRAM + ( 63 << 2 ); pSPRRAM >= SPRRAM; pSPRRAM -= 4 )
    spr.cx = 640;
    for ( pSPRRAM = SPRRAM; pSPRRAM < SPRRAM + 64 * 4; pSPRRAM += 4 )
    {

      spr.cy = 404 + ( pSPRRAM[ SPR_ATTR ] & SPR_ATTR_COLOR );

      spr.x = pSPRRAM[ SPR_X ];
      spr.y = pSPRRAM[ SPR_Y ] + 1;

      spr.attribute = ( ( pSPRRAM[ SPR_ATTR ] << 17 ) & 0x00800000 ) | ( ( pSPRRAM[ SPR_ATTR ] << 15 ) & 0x00400000 );

      if ( PPU_R0 & R0_SP_SIZE )
      {
        //--------------------
        // Sprite size 8x16
        //--------------------
        
        wCHR = ( pSPRRAM[ SPR_CHR ] & 0xfe ) + ( ( pSPRRAM[ SPR_CHR ] & 1 ) << 8 );

        if ( pSPRRAM[ SPR_ATTR ] & SPR_ATTR_V_FLIP )
        {
          ++wCHR;
          spr.u = ( wCHR & 0xf ) << 3;
          spr.v = ( ( wCHR & 0x1f0 ) >> 1 );

          GsSortSprite( &spr, pOT, 0 );

          spr.y += 8;
          spr.u -= 8;
        }
        else
        {
          spr.u = ( wCHR & 0xf ) << 3;
          spr.v = ( ( wCHR & 0x1f0 ) >> 1 );

          GsSortSprite( &spr, pOT, 0 );

          spr.y += 8;
          spr.u += 8;
        }
      }
      else
      {
        //--------------------
        // Sprite size 8x8
        //--------------------
        wCHR = pSPRRAM[ SPR_CHR ];
        spr.u = ( wCHR & 0xf ) << 3;
        spr.v = ( ( wCHR & 0xf0 ) >> 1 ) + bySPBase;
      }
      GsSortSprite( &spr, pOT, 0 );
    }
  }

  byNameTable = PPU_NameTableBank;
  byXScr = PPU_Scr_H_Byte_Next;
  byYScr = PPU_Scr_V_Byte_Next;

  byBGBase = ( ( PPU_R0 & R0_BG_ADDR ) << 3 );

  for ( wY = 1; wY < 30; ++wY )
  {
    wPY = wY + byYScr;
    if ( wPY > 29 )
    {
      // Holizontal Mirror
      byNameTable ^= NAME_TABLE_V_MASK;
      byYScr -= 30;
      wPY = 0;
    }

    wY4 = ( ( wPY & 2 ) << 1 );

    pbyNameTable = PPUBANK[ byNameTable ] + ( ( wY + byYScr ) << 5 ) + byXScr;

    pAttrBase = PPUBANK[ byNameTable ] + 0x3c0 + ( wPY / 4 ) * 8;

    spr.x = -PPU_Scr_H_Bit_Next;
    spr.y = wY * 8 - PPU_Scr_V_Bit_Next;
    for ( wX = 0, wPX = wX + byXScr; wPX < 32; ++wX, ++wPX, spr.x += 8 )
    {
      spr.cy = 400 + ( ( pAttrBase[ wPX >> 2 ] >> ( ( wPX & 2 ) + wY4 ) ) & 3 );
      spr.u = ( *pbyNameTable & 0xf ) << 3;
      spr.v = ( ( *pbyNameTable & 0xf0 ) >> 1 ) + byBGBase;
      GsSortFastSprite( &spr, pOT, 0 );

      ++pbyNameTable;
    }

    // Holizontal Mirror
    byNameTable ^= NAME_TABLE_H_MASK;
    pbyNameTable = PPUBANK[ byNameTable ] + ( wY + byYScr ) * 32;
    pAttrBase = PPUBANK[ byNameTable ] + 0x3c0 + ( wPY / 4 ) * 8;

    spr.x = wX * 8 - PPU_Scr_H_Bit_Next;
    for ( wPX = 0; wX < 33; ++wX, ++wPX, spr.x += 8 )
    {
      spr.cy = 400 + ( ( pAttrBase[ wPX >> 2 ] >> ( ( wPX & 2 ) + wY4 ) ) & 3 );
      spr.u = ( *pbyNameTable & 0xf ) << 3;
      spr.v = ( ( *pbyNameTable & 0xf0 ) >> 1 ) + byBGBase;
      GsSortFastSprite( &spr, pOT, 0 );

      ++pbyNameTable;
    }

    byNameTable ^= NAME_TABLE_H_MASK;

  }

  DrawSync( 0 );
  
  HSTime = VSync( 0 );

  // Switching of the double buffer
  GsSwapDispBuff();

  GsDrawOt( pOT );

}

/*===================================================================*/
/*                                                                   */
/*             pNesX_MemoryCopy() : memcpy                           */
/*                                                                   */
/*===================================================================*/
void *pNesX_MemoryCopy( void *dest, const void *src, int count )
{
/*
 *  memcpy
 *
 *  Parameters
 *    void *dest                       (Write)
 *      Points to the starting address of the copied blockfs destination
 *
 *    const void *src                  (Read)
 *      Points to the starting address of the block of memory to copy
 *
 *    size_t count                     (Read)
 *      Specifies the size, in bytes, of the block of memory to copy
 *
 *  Return values
 *    Pointer of destination
 */

  bcopy( src, dest, count );
  return dest;
}

/*===================================================================*/
/*                                                                   */
/*             pNesX_MemorySet() : Get a joypad state                 */
/*                                                                   */
/*===================================================================*/
void *pNesX_MemorySet( void *dest, int c, int count )
{
/*
 *  memset
 *
 *  Parameters
 *    void *dest                       (Write)
 *      Points to the starting address of the block of memory to fill
 *
 *    int c                            (Read)
 *      Specifies the byte value with which to fill the memory block
 *
 *    size_t count                     (Read)
 *      Specifies the size, in bytes, of the block of memory to fill
 *
 *  Return values
 *    Pointer of destination
 */

  memset( dest, c, count );
  return dest;
}

/*===================================================================*/
/*                                                                   */
/*                DebugPrint() : Print debug message                 */
/*                                                                   */
/*===================================================================*/
void pNesX_DebugPrint( char *pszMsg )
{
}

/*===================================================================*/
/*                                                                   */
/*            pNesX_SetupChr() : Develop character data              */
/*                                                                   */
/*===================================================================*/
void pNesX_SetupChr()
{
/*
 *  Develop character data
 *
 */
 
  BYTE *pbyBGData;
  BYTE byData1;
  BYTE byData2;
  int nIdx;
  int nY;
  int nOff;
  static BYTE *pbyPrevBank[ 8 ];
  int nBank;

  BYTE SprData[ 64 * 32 ];
  RECT rc;
  
  for ( nBank = 0; nBank < 8; ++nBank )
  {
    if ( pbyPrevBank[ nBank ] == PPUBANK[ nBank ] && !( ( ChrBufUpdate >> nBank ) & 1 ) )
      continue;  // Next bank

    /*-------------------------------------------------------------------*/
    /*  An address is different from the last time                       */
    /*    or                                                             */
    /*  An update flag is being set                                      */
    /*-------------------------------------------------------------------*/

    for ( nIdx = 0; nIdx < 64; ++nIdx )
    {
      nOff = ( ( nIdx & 0x30 ) << 5 ) | ( ( nIdx & 0xf ) << 2 );
      
      for ( nY = 0; nY < 8; ++nY )
      {
        pbyBGData = PPUBANK[ nBank ] + ( nIdx << 4 ) + nY;

        byData1 = ( ( pbyBGData[ 0 ] >> 1 ) & 0x55 ) | ( pbyBGData[ 8 ] & 0xAA );
        byData2 = ( pbyBGData[ 0 ] & 0x55 ) | ( ( pbyBGData[ 8 ] << 1 ) & 0xAA );
      
        SprData[ nOff ] = ( ( byData2 >> 2 ) & 0x30 ) | ( ( byData1 >> 6 ) & 3 );
        SprData[ nOff + 1 ] = ( byData2 & 0x30 ) | ( ( byData1 >> 4 ) & 3 );
        SprData[ nOff + 2 ] = ( ( byData2 << 2 ) & 0x30 ) | ( ( byData1 >> 2 ) & 3 );
        SprData[ nOff + 3 ] = ( ( byData2 << 4 ) & 0x30 ) | ( byData1 & 3 );
        
        nOff += 64;
      }
    }

    // Transfer it to the frame buffer
    rc.x = 640 + 256;
    rc.y = 256 + nBank * 32;
    rc.w = 32;
    rc.h = 32;
    LoadImage( &rc, (u_long *)SprData );
    DrawSync( 0 );

    // Keep this address
    pbyPrevBank[ nBank ] = PPUBANK[ nBank ];
  }

  // Reset update flag
  ChrBufUpdate = 0;
}

/*===================================================================*/
/*                                                                   */
/*            pNesX_InitSound() : Initialize Sound                   */
/*                                                                   */
/*===================================================================*/
void pNesX_InitSound()
{
  // initialise the SPU
  SPU_Init();

  // upload the sample
  SPU_UploadData( VAG_SQUARE_ADDR, 0x2000, 25984 );
  SPU_UploadData( VAG_TRIANGLE_ADDR, 0x10000, 25984 );
  SPU_UploadData( VAG_NOISE_ADDR, 0x20000, 25984 );

  // turn it on
  SPU_Enable();
  
  // Tell voice 0 where to find the sample
  SPU_VoiceSetData( 0, 0x2000 );
  SPU_VoiceSetData( 1, 0x2000 );
  SPU_VoiceSetData( 2, 0x10000 );
  SPU_VoiceSetData( 3, 0x20000 );

}

/*===================================================================*/
/*                                                                   */
/*            pNesX_SetFreq() : Set Voice Frequency                  */
/*                                                                   */
/*===================================================================*/
void pNesX_SetFreq( int ch, int freq )
{
  //if ( ch != 0 )
  //  return;
    
  //pitch = freq << 3;

  SPU_VoicePitch( ch, freq << 2 );
  
  SPU_VoiceOn( ch );
}

/*===================================================================*/
/*                                                                   */
/*            pNesX_SetVolume() : Set Voice Volume                   */
/*                                                                   */
/*===================================================================*/
void pNesX_SetVolume( int ch, int vol )
{
  //if ( ch != 0 )
  //  return;

  vol <<= 10;

  SPU_VoiceVol( ch, vol, vol );
}

/*===================================================================*/
/*                                                                   */
/*                     pNesX_Mute() : Mute                           */
/*                                                                   */
/*===================================================================*/
void pNesX_Mute()
{
  SPU_VoiceVol( 0, 0, 0 );
  SPU_VoiceVol( 1, 0, 0 );
  SPU_VoiceVol( 2, 0, 0 );
  SPU_VoiceVol( 3, 0, 0 );
}


